# Ubuntu Developer Summit web application
# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytz
import sys
import time
import urllib2

from datetime import datetime

try:
    from xml.etree import ElementTree
except ImportError:
    from elementtree import ElementTree

from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from django.contrib.auth.models import User

from summit.schedule.fields import NameField

__all__ = (
    'Summit',
)

class SummitManager(models.Manager):
    
    def next(self):
        try:
            return self.order_by('-date_start')[0]
        except IndexError:
            # No summits have been defined in the database yet
            return Summit(name="no_summits", title="No Summits Defined")

class Summit(models.Model):
    STATE_CHOICES = (
        (u'sponsor', u'Sponsorship Requests'),
        (u'schedule', u'Scheduling'),
        (u'public', u'Public'),
    )

    name = NameField(max_length=50)
    title = models.CharField(max_length=100)
    location = models.CharField(max_length=100, blank=True)
    description = models.TextField(max_length=2047, blank=True)
    timezone = models.CharField(max_length=50,
                                choices=[(x, x) for x in pytz.common_timezones])
    last_update = models.DateTimeField(null=True, blank=True)
    state = models.CharField(max_length=10, choices=STATE_CHOICES,
                             default=STATE_CHOICES[0][0])
    date_start = models.DateField(blank=False, null=True)
    date_end = models.DateField(blank=False, null=True)
    
    objects = SummitManager()
    
    class Meta:
        app_label = 'schedule'
        ordering = ('name', )

    def __unicode__(self):
        return self.name

    def localize(self, datetime):
        """Convert a datetime to the summit-local timezone.

        Normally this is used on naive objects from the database which
        contain UTC representations, but it will shift any aware timezone
        properly.
        """
        if datetime.utcoffset() is None:
            datetime = pytz.utc.localize(datetime)
        return datetime.astimezone(pytz.timezone(self.timezone))

    def delocalize(self, datetime):
        """Convert a datetime into a naive UTC representation.

        The returned object is naive and contains the UTC representation
        of the time, which is what we store in the database.  Normally
        this is used on aware objects, if not, the same object is returned.
        """
        if datetime.utcoffset() is not None:
            datetime = datetime.astimezone(pytz.utc)
        return datetime.replace(tzinfo=None)

    @property
    def start(self):
        try:
            return self.slot_set.order_by('start_utc')[0].start
        except IndexError:
            return None

    @property
    def end(self):
        try:
            return self.slot_set.order_by('-end_utc')[0].end
        except IndexError:
            return None

    def dates(self):
        """List of dates for this summit.

        Returns sorted naive date objects in local time.
        """
        dates = set()
        dates.update(slot.start.date() for slot in self.slot_set.all())
        dates.update(slot.end.date() for slot in self.slot_set.all())
        return sorted(dates)

    def public_rooms(self):
        """List of public rooms for this summit.

        Returns sorted rooms.
        """

        def by_title(a, b):
            if a.type == 'plenary':
                return -1
            elif b.type == 'plenary':
                return 1
            elif a.track is None:
                return 1
            elif b.track is None:
                return -1
            else:
                return cmp(a.track.title, b.track.title)

        rooms = self.room_set.exclude(type__exact='closed')
        return sorted(rooms, cmp=by_title)

    def open_rooms(self):
        """List of open rooms for this summit.

        Returns sorted rooms.
        """

        def by_title(a, b):
            if a.track is None:
                return 1
            elif b.track is None:
                return -1
            else:
                return cmp(a.track.title, b.track.title)

        rooms = self.room_set.filter(type__exact='open')
        return sorted(rooms, cmp=by_title)

    def update_from_launchpad(self):
        """Update from Launchpad data."""
        url = ("https://launchpad.net/sprints/%s/+temp-meeting-export"
               % self.name)
        trycounter = 0
        retrytotal = 5
        while trycounter <= retrytotal:
            try:
                 export = urllib2.urlopen(url)
            except urllib2.HTTPError, e:
                trycounter += 1
                if trycounter >= retrytotal:
                    print "Error while calling the launchpad API: " + str(e)
                    sys.exit(1)
                else: 
                    time.sleep(2)
            else:
                break
        tree = ElementTree.parse(export)

        for elem in tree.find("attendees").findall("person"):
            self.update_attendee_from_launchpad(elem)
        for elem in tree.find("unscheduled").findall("meeting"):
            self.update_meeting_from_launchpad(elem)

        in_lp = set(self.meeting_set.filter(name__exact=m.get("name")[:50])[0] for m in tree.find("unscheduled").findall("meeting"))
        in_db = set(m for m in self.meeting_set.exclude(spec_url=''))

        for extra in in_db.difference(in_lp):
            if not len(extra.agenda_set.all()):
                print "will delete %s" % extra.name
                extra.delete()

        self.last_update = datetime.utcnow()
        self.save()

    def update_attendee_from_launchpad(self, elem):
        """Update or create single attendee from Launchpad data."""
        username = elem.get("name", "")
        if not username:
            return

        print "user %s" % username

        try:
            attendee = self.attendee_set.get(user__username__exact=username[:30])
        except ObjectDoesNotExist:
            try:
                user = User.objects.get(username__exact=username[:30])
            except ObjectDoesNotExist:
                user = User.objects.create_user(username[:30], '', password=None)

            # Create with any start/end time since we overwrite shortly
            attendee = self.attendee_set.create(user=user,
                                                start=datetime.utcnow(),
                                                end=datetime.utcnow())

        attendee.update_from_launchpad(elem)

    def update_meeting_from_launchpad(self, elem):
        """Update or create single meeting from Launchpad data."""
        name = elem.get("name", "")
        if not name:
            return
        name = name.replace('.','-')

        print "meeting %s" % name
        meeting = ""
        try:
            meeting = self.meeting_set.get(name__exact=name[:50])
        except ObjectDoesNotExist:
            meeting = self.meeting_set.create(name=name[:50], title=name[:100])
        except:
            pass

        if meeting:
            meeting.update_from_launchpad(elem)

    def fill_schedule(self):
        """Fill empty slots and rooms in the schedule with unscheduled meetings

        This is a pretty simple best-fit/first-come-first-served scheduler,
        but it suffices.
        """
        for meeting in self.meeting_set.filter(agenda__isnull=True):
            meeting.try_schedule()

    def check_schedule(self):
        """Check the schedule for existant errors."""
        for meeting in self.meeting_set.all():
            for agenda in meeting.agenda_set.all():
                try:
                    missing = meeting.check_schedule(agenda.slot, agenda.room)
                    if len(missing):
                        print "Warning: required people not available: %s at %s in %s: %s" % (
                            meeting, agenda.slot, agenda.room,
                            ', '.join(m.user.username for m in missing))
                except meeting.SchedulingError, e:
                    print "Error: %s at %s in %s: %s" % (meeting,
                                                         agenda.slot,
                                                         agenda.room,
                                                         e)
