# Ubuntu Developer Summit web application
# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from time import strftime
import datetime

from django.conf import settings
from django.contrib.auth import logout
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render_to_response, get_object_or_404
from django.template import RequestContext

from summit.schedule.decorators import (summit_required,
                                        summit_only_required,
                                        summit_attendee_required)

from summit.schedule.models import Summit, Slot

from summit.schedule.forms import AttendeeInterestForm

from summit.schedule.render import schedule_factory, Schedule

__all__ = (
    'index',
    'summit',
    'interest',
    'by_date',
    'by_room',
    'csv',
    'today_view',
)


def index(request):
    summit = Summit.objects.next()

    context = {
        'summit': summit,
    }
    return render_to_response("schedule/index.html", context,
                              context_instance=RequestContext(request))


@summit_required
def summit(request, summit, attendee):
    edit = False
    if (summit.state != 'public'
        and 'readonly' not in request.GET) \
            or 'edit' in request.GET:
        if request.user.is_authenticated() \
                and request.user.has_perm('schedule.change_agenda'):
            edit = True
    tracks = summit.track_set.all()
    if settings.LINARO_ONLY:
        tracks = tracks.filter(slug__contains='linaro')

    context = {
        'summit': summit,
        'attendee': attendee,
        'edit': edit,
        'tracks': tracks,
    }
    return render_to_response("schedule/summit.html", context,
                              context_instance=RequestContext(request))


@summit_attendee_required
def interest(request, summit, attendee):
    if request.method == 'POST':
        form = AttendeeInterestForm(request.POST, instance=attendee)
        if form.is_valid():
            form.save()
            request.user.message_set.create(
                message="Your interest has been updated.")
            return HttpResponseRedirect("/%s" % summit.name)
    else:
        form = AttendeeInterestForm(instance=attendee)

    context = {
        'summit': summit,
        'attendee': attendee,
        'form': form,
    }
    return render_to_response("schedule/interest.html", context,
                              context_instance=RequestContext(request))


@summit_required
def by_date(request, summit, attendee, date):
    schedule = schedule_factory(request, summit, attendee, date=date)

    if request.method == 'POST':
        return schedule.save_change()
    else:
        viewdate = datetime.datetime.strptime(date, "%Y-%m-%d")
        day = datetime.timedelta(days=1)
		
        context = {
            'summit': summit,
            'attendee': attendee,
            'schedule': schedule,
            'ical': '/%s.ical' % summit.name,
            'autoreload': 'reload' in request.GET,
            'nextday': viewdate + day,
            'previousday': viewdate - day,
        }
        converted_date = summit.delocalize(datetime.datetime.strptime(date, "%Y-%m-%d"))
        if Slot.objects.filter(summit=summit, start_utc__gte=converted_date, end_utc__lte=converted_date+datetime.timedelta(days=1)).count() > 0:
            schedule.calculate()
        else:
            return render_to_response("schedule/nosession.html", context,
                              context_instance=RequestContext(request))
    return render_to_response("schedule/schedule.html", context,
                              context_instance=RequestContext(request))

def today_view(request):
    summit = Summit.objects.next()
    today = summit.localize(datetime.datetime.now()).date()
    return by_date(request, summit.name, today.strftime("%Y-%m-%d"))

@summit_required
def next_session(request, summit, attendee):
    schedule = Schedule(request, summit, attendee)
    schedule.nextonly = True

    schedule.calculate()

    context = {
        'summit': summit,
        'attendee': attendee,
        'schedule': schedule,
        'autoreload': 'reload' in request.GET,
    }
    return render_to_response("schedule/nextsession.html", context,
                              context_instance=RequestContext(request))

@summit_required
def by_room(request, summit, attendee, room_name):
    try:
        room = summit.room_set.get(tracks__title__iexact=room_name)
    except (ObjectDoesNotExist, MultipleObjectsReturned):
        room = get_object_or_404(summit.room_set, name__exact=room_name)

    schedule = schedule_factory(request, summit, attendee, room=room)

    if request.method == 'POST':
        return schedule.save_change()
    else:
        schedule.calculate()

    context = {
        'summit': summit,
        'attendee': attendee,
        'schedule': schedule,
        'ical': '/%s/room/%s.ical' % (summit.name, room.name),
        'autoreload': 'reload' in request.GET,
    }
    return render_to_response("schedule/schedule.html", context,
                              context_instance=RequestContext(request))


@summit_required
def by_track(request, summit, attendee, track_slug):
    track = get_object_or_404(summit.track_set, slug__iexact=track_slug)

    schedule = Schedule(request, summit, attendee, track=track)

    if request.method == 'POST':
        return schedule.save_change()
    else:
        schedule.calculate()

    context = {
        'summit': summit,
        'attendee': attendee,
        'schedule': schedule,
        'autoreload': 'reload' in request.GET,
        'ical': '/%s/track/%s.ical' % (summit.name, track.slug),
    }
    return render_to_response("schedule/schedule.html", context,
                              context_instance=RequestContext(request))

@summit_required
def meeting(request, summit, attendee, meeting_slug):
    meeting = get_object_or_404(summit.meeting_set, name__iexact=meeting_slug)
    agendaitems=meeting.agenda_set.all()
    participants=meeting.participant_set.all()
    tracks=meeting.tracks.all()
    context = {
        'summit': summit,
        'meeting':meeting,
        'agenda_items':agendaitems,
        'participants':participants,
        'tracks':tracks,
        'ETHERPAD_HOST':settings.ETHERPAD_HOST
    }
    return render_to_response("schedule/meeting.html", context,
                              context_instance=RequestContext(request))


@summit_only_required
def csv(request, summit):
    schedule = Schedule(request, summit)
    schedule.calculate()

    return HttpResponse(schedule.as_csv(),
                        mimetype='text/csv')

@summit_only_required
def ical(request, summit):
    """Return any list events as an ical"""
    schedule = Schedule(request, summit)
    schedule.calculate()
    
    filename = "%s.ical" % summit.name.replace(' ', '-').lower()
    response = HttpResponse(mimetype='text/calendar')
    response['Content-Disposition'] = 'attachment; filename=%s' % filename.encode('ascii', 'replace')
    response.write(schedule.as_ical())
    return response

@summit_only_required
def user_ical(request, summit, username):
    """Returns a user's registered events as an ical"""
    schedule = Schedule(request, summit)
    schedule.calculate()
    
    filename = "%s_%s.ical" % (summit.name, username)
    response = HttpResponse(mimetype='text/calendar')
    response['Content-Disposition'] = 'attachment; filename=%s' % filename.replace(' ', '-').lower().encode('ascii', 'replace')
    response.write(schedule.as_ical(only_username=username, show_private=False))
    return response
    
@summit_only_required
def room_ical(request, summit, room_name):
    """Returns a room's events as an ical"""
    schedule = Schedule(request, summit)
    schedule.calculate()
    
    filename = "%s_%s.ical" % (summit.name, room_name)
    response = HttpResponse(mimetype='text/calendar')
    response['Content-Disposition'] = 'attachment; filename=%s' % filename.replace(' ', '-').lower().encode('ascii', 'replace')
    response.write(schedule.as_ical(only_room=room_name))
    return response
    
@summit_only_required
def track_ical(request, summit, track_slug):
    """Returns a track's events as an ical"""
    schedule = Schedule(request, summit)
    schedule.calculate()
    
    filename = "%s_%s.ical" % (summit.name, track_slug)
    response = HttpResponse(mimetype='text/calendar')
    response['Content-Disposition'] = 'attachment; filename=%s' % filename.replace(' ', '-').lower().encode('ascii', 'replace')
    response.write(schedule.as_ical(only_track=track_slug))
    return response
    
def logout_view(request):
    logout(request)
    return HttpResponseRedirect('/')

def mobile(request):
    return render_to_response("schedule/mobile.html", {}, RequestContext(request))

def past(request):
    pastsummit = Summit.objects.filter(date_end__lte=datetime.date.today())
    context = {
        'past_summit': pastsummit,
    }
    return render_to_response("schedule/past_summit.html", context,
                          context_instance=RequestContext(request))
